-------------------------------------------------------------- 
--  Subject : Theorem proving ( module )
--  Compare with Appendix A in "Chang C.L and Lee R.C Symbolic 
--               Logic and Mechanical Theorem Proving", 
--               Academic Press, New York, 1973                            
--  Copyright (c) 1999 by A.Baranovsky
--------------------------------------------------------------

uses sets, conv

  constant
    ArrConst = [  '$1',  '$2',  '$3',  '$4',  '$5',  '$6',  '$7',  '$8' ]     
    AX       = [ '#X1', '#X2', '#X3', '#X4', '#X5', '#X6', '#X7', '#X8' ] 
    AY       = [ '#Y1', '#Y2', '#Y3', '#Y4', '#Y5', '#Y6', '#Y7', '#Y8' ] 
  end 

  function IsLogVar( X ) inline
  return ( ClassOf( X ) = ClassString ) and ( Substr( X, 1, 1 ) = '#' )

  class Clause : Array
    function InitClause( L ) constructor          -- Create clause
    function Resolv( A, B, LA, LB ) constructor   -- Create a resolvent
    function Rename( L )                          -- Rename clause L
    function ConstRename()          -- Rename clause L by the constants
    function GetNX()                -- Return father's number
    function GetNY()                -- Return mother's number
    function GetNumb()              -- Return given clause's number
    function GetCount()             -- Return number of literals except ANS
  end

  implementation Clause

    local Param, NParam, Numb, NX, NY, ND, shared Num   -- object's fields

    function Rename( L ) 
      for I:=1 to NParam do
        Destroy( Param[ I ] )             -- Destroy I-th element of Param
        Param[ I ] := L[ I ]              -- Assign I-th element of Param
      end
    return 0

    function ConstRename() 
      for I:=1 to NParam do
        Destroy( Param[ I ] )            -- Destroy Param[ I ] 
        Param[ I ] == ArrConst[ I ]      -- Identify Param[ I ] 
      end                                -- with ArrConst[ I ]
    return 0

    function Print()
      writeln( '(' + STR(NX) + ',' + STR(NY) + ',' + STR(ND) + ')' ) 
      writeln( STR( Numb ) + ') ' + ConvToStr( Self() ) )
--    writeln( 'Parameters '+ Str( Param ) )
    return 0

    function Copy( var A )              -- Literal copying ( private method )
      N := Card( A )
      R := array( N )                   -- Create array R
      for I:=1 to N do
        AI == A[ I ]                    -- Let AI is A[ I ]
        RI == R[ I ]                    -- Let RI is R[ I ]
        if    ClassOf( AI ) = ClassArray then
          RI := Copy( AI )              -- Recursive call
        elsif IsLogVar( AI ) then
          Exist := FALSE
          for J:=1 to NParam do         -- Search A[ I ] in Param[ I ]
            if Param[ J ] = AI then     -- Found ?
              Exist := TRUE
              @RI == Param[ J ]         --   Identify R[ I ] with Param[ J ]
            end
          end
          if not Exist then             -- If not found then
            Inc( NParam )               -- increase number of log-variable    
            Param[ NParam ] := AI
            @RI == Param[ NParam ]
          end
        else                         
          @RI == AI                  -- Identify R[ I ] with A[ I ]
        end
      end
    return R

    function GetCount()    
                           -- ( ANS-predicates are in a tail )
      P == Self()                                             
      N := Card( P )
      while ( N > 0 ) and ( P[ N, 2 ] = 'ans' ) do               
        Dec( N )
      end
    return N

    function GetNumb()  inline  
    return Numb

    function GetNX()  inline   
    return NX

    function GetNY()  inline    
    return NY

    function Resolv( InitA, InitB, LA, LB )
      object &A, &B : Clause         -- Pseudonym declarations
      A == InitA
      B == InitB

      N1 := Card( A )
      N2 := Card( B )
      N  := N1 + N2 - 2
      Array.Init( N )               -- Call from ancestor class
      Param  := array( 8 )
      NParam := 0
      P == Self()
      J := 0

      -- Copying A's literals -----               
     
      for I:=1 to A.GetCount() do                                        
        if I <> LA then
          Inc( J )
          P[ J ] := Copy( A[ I ] )
        end
      end

      -- Copying B's literals -----                

      for I:=1 to B.GetCount() do  
        if I <> LB then
          Inc( J )
          P[ J ] := Copy( B[ I ] )
        end
      end

     ------ Copying ANS-predicates from A ---------

      while ( N1 > 0 ) and ( A[ N1, 2 ] = 'ans' ) do 
        Inc( J ) 
        P[ J ] := Copy( A[ N1 ] )
        Dec( N1 )
      end

     ------ Copying ANS-predicates from B --------- 

      while ( N2 > 0 ) and ( B[ N2, 2 ] = 'ans' ) do
        Inc( J )
        P[ J ] := Copy( B[ N2 ] )
        Dec( N2 )
      end

      Inc( Num )
      Numb := Num
      NX := A.GetNumb()  
      NY := B.GetNumb()
      ND := LB
    return 0

    function InitClause( var L ) 
      object &PL : Definition       -- Pseudonym declaration
      PL == L

      T := EvalObject( L, [] )      -- Evaluate lambda-object

      Array.Init( Card( T ) )       -- Call from ancestor class 

      P == Self()
      for I:=1 to Card( T ) do
        P[ I ] << T[ I ]            -- Replace P components by T components
      end

      P  == PL.GetParam() 

      if not Null( P ) then
        Param << P
      else
        Param := []
      end

      NParam := Card( Param ) 

      if Null( Num ) then 
        Num := 0
      end
      Inc( Num )
      Numb := Num

      NX := 1000
      NY := 0
      ND := 0
    return 0
  end  -- of Clause implementation

  function Inside( Key, T )           -- Entrance check
    if ClassOf( T ) = ClassArray then
      Found := FALSE
      I := 2
      K := Card( T )
      while ( I <= K ) and ( not Found ) do
        A == T[ I ]
        CA := ClassOf( A )
        if CA = ClassArray then 
          Found := Inside( Key, A )
        elsif IsLogVar( A ) and ( A = Key ) then   
          Found := TRUE
        end
        Inc( I )
      end
    elsif not IsLogVar( T ) then 
      Found := FALSE
    else
      Found := Key = T
    end    
  return Found

  function Unify( N, T1, T2 )  
    Failrue := FALSE
    I := N 
    K := Card( T1 ) 

    ---- Synchronous postorder traversal of T1 and T2 ------

    while ( I <= K ) and ( not Failrue ) do
      P1 == T1[ I ]               -- Let P1 is T1[ I ]
      P2 == T2[ I ]               -- Let P2 is T2[ I ]
      C1 := ClassOf( P1 )         -- Find out a type of T1[ I ]
      C2 := ClassOf( P2 )         -- Find out a type of T2[ I ]
      if ( C1 = ClassArray ) and ( C2 = ClassArray ) then 
        Failrue := Unify( 1, P1, P2 )    -- Recursive call
      elsif IsLogVar( P1 ) and ( not Inside( P1, P2 ) ) then 
        Destroy( @P1 )            -- Destroy log-variable ( from Param )
        @P1 == P2                 -- Identify log-variable with T2[ I ]
      elsif IsLogVar( P2 ) and ( not Inside( P2, P1 ) ) then  
        Destroy( @P2 )            -- Destroy log-variable ( from Param ) 
        @P2 == P1                 -- Identify log-variable with T1[ I ]
      elsif (( C1 <> C2  ) or ( P1 <> P2 )) then 
        Failrue := TRUE
      end
      Inc( I )
    end
  return Failrue


  function Contradict( InitU, InitV )  -- Search of the direct contradiction  

    object &U, &V : Collection         -- Pseudonym declarations
    U == InitU
    V == InitV

    object &UI, &VJ : Clause           -- Pseudonym declarations 

    for I:=1 to U.GetCount() do
      UI == U[ I ] 
      UIF := UI[ 1, 2 ]
      for J:=1 to V.GetCount() do
        VJ == V[ J ]
        if UIF = VJ[ 1, 2 ] then
          UI.Rename( AX )             -- Rename before unification
          VJ.Rename( AY ) 
          F := Unify( 3, UI[ 1 ], VJ[ 1 ] )
          if F = 0 then
            R := [ UI.GetNumb(), VJ.GetNumb() ]   
            goto Fin
          end
        end
      end
    end
    R := [] 
    label Fin
  return R

  function STest( InitS, InitU )  -- Subsumption check

    object &S : Collection        -- Pseudonym declarations  
    object &SI, &U : Clause

    S == InitS
    U == InitU

    UF := U[ 1, 2 ]

    R := FALSE

    for I:=1 to S.GetCount() do
      SI == S[ I ]
      if UF = SI[ 1, 2 ] then
        SI.ConstRename()         -- Rename by the ArrConst
        U.Rename( AX )   
        F := Unify( 3, SI[ 1 ], U[ 1 ] )
        if F = 0 then
          R := TRUE   
          goto Fin
        end
      end
    end
    label Fin
  return R

  function GUnit( var InitS1, var InitS2, var InitW, InitC, var A, var N3, var HIST, var HC )
    object &S1, &S2, &W, &P, WW, POS(), NEG() : Collection
    object &U, C, R : Clause

    S1 == InitS1
    S2 == InitS2
    W  == InitW

    WW := S1 + S2

    Stack := array( 100 )          -- Create stack
    Stack[ 1 ] := InitC
    NS := 1                        -- Level of the Stack

    ---- Postorder traversal of deduction tree

    repeat
      C := Stack[ NS ]
      Dec( NS )

      for I:=1 to W.GetCount() do       -- Cycle by elements of support set W
        U == W[ I ]
        if U.GetNumb() < C.GetNX() then
          SU := U[ 1, 1 ]               -- Sign
          FU := U[ 1, 2 ]               -- Functor

          for J:=1 to C.GetCount() do   -- Cycle by non-ANS literals of C

            if ( SU <> C[ J, 1 ] ) and ( FU = C[ J, 2 ] ) then

              U.Rename( AX )   -- Rename before unification
              C.Rename( AY )

--            writeln( 'Source clauses ' + STR( J ) )
--            U.Print()
--            C.Print()

              Failrue := Unify( 3, U[ 1 ], C[ J ] )  -- Unification

              if not Failrue then
                R := Clause( U, C, 1, J )

                writeln( 'Resolvent' )          
                R.Rename( AX )
                R.Print()

                Inc( HC )           -- App to the history
                HIST[ HC ] := [ U.GetNumb(), C.GetNumb(), J, * ]

                if R.GetCount() = 1 then     -- if R is unit clause
                  if R[ 1, 1 ] = '+' then
                    S == S1
                    P == POS
                  else
                    S == S2
                    P == NEG
                  end
                  if ( not STest( S, R ) ) and ( not STest( P, R ) ) then
                    Inc( N3 )                -- Save R into output array A
                    A[ N3 ] << R
                    P.App( A[ N3 ] )
                  end
                else                         -- If R is non-unit clause then
                  Inc( NS )                  --     Stack <--- R      
                  Stack[ NS ] := R
                end
              end
            end
          end
        end
      end
      W == WW
    until NS = 0

    Z := Contradict( S1, NEG )   -- Search direct contradiction
    if Card( Z ) = 0 then
      Z := Contradict( S2, POS ) 
      if Card( Z ) = 0 then
        S1 := S1 + POS
        S2 := S2 + NEG
      end 
    end
  return Z

  function MarkTree( I, var HIST, N1 )   -- Pick out a deduction tree from HIST  
    if I > N1 then                         
      HIST[ I, 4 ] := TRUE
      MarkTree( HIST[ I, 1 ], HIST, N1 )
      MarkTree( HIST[ I, 2 ], HIST, N1 ) 
    end 
  return 0

-------- Main program of TPU algorithm -----------------

  function TPU( var A, InitW, N1, N2 )
    object S1(), S2(), S3(), &S, &WK, &WJ : Collection
    object &CK, &U, &V, &AI : Clause 

    HIST := array( 250 )       -- Create the history array HIST
    HC   := N1

    N3 := N1

    W := array( N1 )           -- Create the support set W

    ----- Create positive unit clause set S1
    ----- Create negative unit clause set S2
    ----- Create non-unit clause set S3
    ----- ( S1, S2, S3 consist of peudonymes of array elements of A  )

    for I:=1 to N1 do
      AI == A[ I ] 
      if AI.GetCount() = 1 then
        if AI[ 1 ][ 1 ] = '+' then
          S == S1
        else
          S == S2
        end
      else
        S == S3
      end
      S.App( AI )
      WI == InitW[ I ]
      if not Null( WI ) then
        K := S3.GetCount()
        WK == W[ K ]
        WK := Collection() 
        for J:=1 to Card( WI ) do
          WK.App( A[ WI[ J ] ] )
        end        
      end
    end

 -- Search direct contradiction between S1 and S2

    R := Contradict( S1, S2 )                                  
    if Card( R ) > 0 then
      goto Fin
    end 

    K := 1

    for I:=1 to N2 do           -- Repeat N2 times

      CK == S3[ K ] 
      WK == W[ K ] 
      N := N3

      R := GUnit( S1, S2, WK, CK, A, N3, HIST, HC )

      if Card( R ) > 0 then 
        goto Fin                 -- The solution is found
      end 

      -- Rebuild support sets ---------

      if N3 > N then
        for J:=1 to S3.GetCount() do
          if J = K then
            Destroy( W[ J ] )
          end
          WJ == W[ J ]  
          if NULL( WJ ) then
            WJ := Collection()
          end
          for L:=(N+1) to N3 do
            AI == A[ L ]
            if AI.GetCount() = 1 then
              WJ.App( AI )
            end
          end
        end
      end

      Inc( K )
      if K > S3.GetCount() then
        K := 1      -- Begin at first
      end
    end
    
    label Fin 

    ------- Output  ---------------------

--    writeln( '' )
--    writeln( R )
--    writeln( '' )

    if Card( R ) = 0 then
      writeln( 'Not proved' )          
    else
      MarkTree( R[1], HIST, N1 )
      MarkTree( R[2], HIST, N1 ) 
      writeln( 'The proof :' ) 
      for I:=N1+1 to HC do
        HI == HIST[ I ]
        if not Null( HI[4] ) then
          writeln( STR(I)+' '+STR( HI[1] )+' '+STR( HI[2] )+' '+STR( HI[3] ) )         
        end  
      end
      writeln( 'Contradiction ' + STR( R[1] ) + ' ' + STR( R[2] ) )
    end
  return 0

return



