-------------------------------------------------------------- 
--  Subject : Corutines
--  Copyright (c) 1999 by A.Baranovsky
--------------------------------------------------------------

  constant 
    Key   = 1
    Value = 2
  end

  class TAgent
    function Init( InitA, InitSolver )
    function EOF()
    function ReadRec( var R ) 
    function WriteRec( R )
    function Send( var R )
    function Receive( R )
  end

  class TSolver
    function Init( iAgent1, iAgent2, iAgent3 )
    function Check()
  end

---------------- TAgent ----------------------------------------------

  implementation TAgent

    local A, RecNo, Solver, Prev

    function Init( InitA, InitSolver )
      object &A      : array
      object &Solver : TSolver 
      A      == InitA 
      Solver == InitSolver            
      RecNo := 1
    return NIL

    function EOF()
    return RecNo > Card( A )

    function ReadRec( var R )
      R     := A[ RecNo ] 
      RecNo := RecNo + 1
    return NIL

    function WriteRec( R )
      writeln( STR( RecNo ) + ' '+ STR( R[ Key ] ) + ' ' + R[ Value ] )

      A[ RecNo ] := R
      RecNo := RecNo + 1
    return NIL

    function Send( var R )
      if not EOF() then
        ReadRec( R )
      else
        Destroy( @R )
      end
    return NIL

    function Receive( R )
      repeat
        if not NULL( R ) then
          Prev := R
        end

        resume Solver.Check()
          
        while ( not NULL( R ) ) and ( Prev[ Key ] = R[ Key ] ) do
          Prev[ Value ] := Prev[ Value ] + R[ Value ]
          resume Solver.Check()
        end

        WriteRec( Prev )

      until false
    return NIL
  end

  ----------------- TSolver --------------------------------

  implementation TSolver

    local Agent1, Agent2, Agent3, R1, R2

    function Init( iAgent1, iAgent2, iAgent3 )
      object &Agent1, &Agent2, &Agent3 : TAgent
      Agent1 == iAgent1
      Agent2 == iAgent2
      Agent3 == iAgent3
    return NIL

    function Check()
      Agent1.Send( R1 )
      Agent2.Send( R2 )
      while ( not NULL( R1 ) ) or ( not NULL( R2 ) ) do  
        if    NULL( R1 ) then                              
          resume Agent3.Receive( R2 )
          Agent2.Send( R2 )
        elsif NULL( R2 ) then
          resume Agent3.Receive( R1 )
          Agent1.Send( R1 )
        elsif R2[ Key ] < R1[ Key ] then
          resume Agent3.Receive( R2 )
          Agent2.Send( R2 )
        else
          resume Agent3.Receive( R1 )
          Agent1.Send( R1 )
        end
      end
      resume Agent3.Receive( Dummy )
    return NIL
  end

  function CreateArray( var A, N, M )
    A := array( N ) 
    for I:=1 to N do
      A[ I ] := array( M ) 
    end
  return 0

  function WriteArray( A )
    writeln( '' )
    for I:=1 to Card( A ) do
      writeln( STR( A[ I, Key ] ) + ' ' + A[ I, Value ] )
    end   
  return 0

  -------------- Main program --------------------- 

  CreateArray( F3, 10, 2 ) 

  F1 := [ [5,'U1'], [7,'U2'], [7,'U3'], [8,'U4'] ]
  F2 := [ [3,'V1'], [7,'V2'], [8,'V3'] ] 

  WriteArray( F1 )
  WriteArray( F2 )

  object Agent1, Agent2, Agent3 : TAgent
  object Solver : TSolver

  Agent1.Init( F1, Solver )
  Agent2.Init( F2, Solver )
  Agent3.Init( F3, Solver )
  Solver.Init( Agent1, Agent2, Agent3 )

  writeln( '' )

  Solver.Check()

  return

