--------------------------------------------------------------  
--  Subject : Linear algebra, Gauss' method
--  Copyright (c) 1999 by A.Baranovsky
--------------------------------------------------------------
constant
  EPS = 0.00000001
end

function Swap( var A, var B )
  X << A
  A << B
  B << X
return 0

function Gauss( var A, var B, var X, N )
  err := FALSE    
  for I:=1 to N-1 do
    Big := Abs( A[ I, I ] )
    L   := I
    for j :=I + 1 to N do
      temp := Abs( A[ J, I ] )
      if temp > Big then   
        Big := temp
        L   := J
      end
    end

    if Big < EPS then 
      err := TRUE
    else
      if L <> I then
        Swap( A[ L ], A[ I ] )       -- Swapping of lines
        Swap( B[ L ], B[ I ] )       -- Swapping of elements
      end
      for J :=I + 1 to N do
        temp := A[ J, I ] / A[ I, I ]
        for K:=1 to N do
          A[ J, K ] := A[ J, K ] - temp * A[ I, K ]
        end 
        B[ J ] := B[ J ] - temp * B[ I ]
      end
    end 
  end

  if Abs( A[ N, N ] ) < EPS then 
    err := TRUE
  else
    X[ N ] := B[ N ] / A[ N, N ]
    I := N - 1
    repeat
      sum := 0.0
      for J:= I + 1 to N do
        sum := sum + A[ I, J ] * X[ J ]
      end
      X[ I ] := ( B[ I ] - sum ) / A[ I , I ]
      I := I - 1
    until I = 0
  end
return Err

function Init( var A, var B, var X, N ) 
  A := array( N )
  B := array( N )
  X := array( N ) 
  for I:=1 to N do
    B[ I ] := Random( 100 ) + 1.0
    X[ I ] := 0.0
    A[ I ] := array( N )
    for J:=1 to N do
      A[ I, J ] := Random( 100 ) + 1.0
    end
  end
return 0

function Show( var A, var B, var X, N )  
  writeln( '' )
  for I:=1 to N do
    R := 0.0
    for J:=1 to N do
      R := R + A[ I, J ] * X[ J ]
    end
    writeln( A[ I ] )
    writeln( R )
  end
  writeln( B ) 
  writeln( X )
  writeln( '' )
  writeln( '' )
return 0

----------------- Main program ----------------

  N := 20

  Init( A, B, X, N )
  Show( A, B, X, N )
  Err := Gauss( A, B, X, N )
  Show( A, B, X, N )

return
