-------------------------------------------------------------- 
--  System module 
--  Copyright (c) 1999 by A.Baranovsky
--------------------------------------------------------------

  function AliasOf( X ) external
  function ClassOf( X ) external
  function Self() external
  function Random( X ) external
  function Card( X ) external
  function Ord( X ) external
  function Chr( X ) external
  function Null( X ) external
  function Destroy( X ) external
  function Str( X ) external
  function Volume( X ) external
  function Double( X, Y ) external
  function EvalExpr( X ) external
  function EvalLine( X ) external
  function EvalObject( X, A ) external
  function GetError() external
  function IsConst( X ) external
  function Inheritor( X, Y ) external
  function Inc( X ) external
  function Dec( X ) external
  function writeln( S ) external
  function MemAvail() external
  function MaxAvail() external
  function Substr( S, I, L ) external
  function NameOf( X ) external
  function ShowMessage( S ) external

  class Array
    function Init( N ) constructor
    function "[]"( N )
  end

  implementation Array
    local A
    function Init( N ) external
    function "[]"( N ) inline
    return AliasOf( A[ N ] )
  end

  class Number
    function Put( X ) constructor
    function ":="( X )
    function "-1"( X )
  end

  implementation Number

    local A

    function Put( X )
      A := X
    return 0

    function PutValue( X ) external

    function ":="( X )
      PutValue( X )
    return A

    function "-1"( X )
      PutValue( X )
      A := -A
    return A
  end

  class Integer : Number
    function "+"( X, Y )
    function "-"( X, Y )
    function "*"( X, Y )
    function "/"( X, Y )
    function "<"( Z )
    function ">"( Z )
    function ">="( Z )
    function "<="( Z )
    function "="( Z ) 
    function "<>"( Z )
  end

  implementation Integer

    function Add( X, Y ) external
    function Sub( X, Y ) external
    function Mult( X, Y ) external
    function Div( X, Y ) external

    function "*"( X, Y )
    return ":="( Mult( X, Y ) )

    function "+"( X, Y )
    return ":="( Add( X, Y ) )

    function "-"( X, Y )
    return ":="( Sub( X, Y ) )

    function "/"( X, Y )
    return ":="( Div( X, Y ) )

    function "<"( Z )  external
    function ">"( Z )  external
    function ">="( Z ) external
    function "<="( Z ) external
    function "="( Z )  external
    function "<>"( Z ) external
  end

  class Real : Number
    function "integer"( X )

    function "+"( X, Y )
    function "-"( X, Y )
    function "*"( X, Y )
    function "/"( X, Y )
    function "<"( Z )
    function ">"( Z )
    function ">="( Z )
    function "<="( Z )
    function "="( Z )
    function "<>"( Z )
  end

  implementation Real
    function "integer"( X )
      ":="( X )
    return Self()

    function Add( X, Y ) external
    function Sub( X, Y ) external
    function Mult( X, Y ) external
    function Div( X, Y ) external

    function "*"( X, Y )
    return ":="( Mult( X, Y ) )

    function "+"( X, Y )
    return ":="( Add( X, Y ) )

    function "-"( X, Y )
    return ":="( Sub( X, Y ) )

    function "/"( X, Y )
    return ":="( Div( X, Y ) )

    function "<"( Z )  external
    function ">"( Z )  external
    function ">="( Z ) external
    function "<="( Z ) external
    function "="( Z )  external
    function "<>"( Z ) external
  end

  class String
    function ":="( X )
    function "+"( X, Y )
    function "<"( Z )
    function ">"( Z )
    function ">="( Z )
    function "<="( Z )
    function "="( Z )
    function "<>"( Z )
  end

  implementation String
    function ":="( X ) external
    function "+"( X, Y ) external
    function "<"( Z ) external
    function ">"( Z ) external
    function ">="( Z ) external
    function "<="( Z ) external
    function "="( Z )  external
    function "<>"( Z ) external
  end

  class Definition
    function GetParam()
  end

  implementation Definition
    local Param, Context, A
    function GetParam()
    return AliasOf( Param )
  end

  class Exception
    function Create() constructor
    function Show()
  end

  function EOL()
  return Chr( 13 ) + Chr( 10 )

  implementation Exception
    local PLine, Code, Title, File, NLine, SLine
    function Create() external
    function Show()
      S := 'Error :' + EOL()
      S := S + '_____________________________________' + EOL()
      S := S + 'File '  + File  + EOL()
      S := S + 'Line '  + STR( NLine )  + EOL()
      S := S + SLine  + EOL()
      S := S + Title  + EOL()
      S := S + 'PLine ' + STR( PLine ) + EOL()
      S := S + 'Code  ' + STR( Code ) + EOL()
      ShowMessage( S )
    return 0
  end


  class EUndefError : Exception
  end

  class EUnary : Exception
  end

  class EConversion : Exception
  end

  class EBinary : Exception
  end

  class ETypeMismatch : Exception
  end

  class EAssignment : Exception
  end

  class EFunctionCall : Exception
  end

  class EDivByZero : Exception
  end

  class ERangeError : Exception
  end

  class EPatternMismatch : Exception
  end

  class EUndefExpected : Exception
  end

  class EExternalError : Exception
  end

  class EHeapOverflow : Exception
  end

  class EParameterPassing : Exception
  end

  function ErrorHandler( InitE )
    if Inheritor( InitE, Exception() ) then
      object &E : Exception
      E == InitE
      E.Show()
    else
      ShowMessage( 'Unhandled exception' )
    end
  return 0

  constant
    ClassArray      = ClassOf( [] )
    ClassInteger    = ClassOf( 1 )
    ClassReal       = ClassOf( 1.0 )
    ClassString     = ClassOf( '' )
    true            = 1
    false           = 0
    NIL             = 0
  end

  function Abs( X )
    if X > 0 then
      R := X
    else
      R := -X
    end
  return R

return
